# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------


from collections import namedtuple
# pylint: disable=no-name-in-module
from distutils.spawn import find_executable
from enum import Enum
from pathlib import Path

import glob
import json
import os
import re
import shlex
import subprocess
import sys
import tempfile
import traceback
from typing import Dict, List, Optional

from codechecker_analyzer.analyzers import clangsa

from codechecker_common.logger import get_logger
from codechecker_common.util import load_json

from .. import gcc_toolchain
from .build_action import BuildAction

LOG = get_logger('buildlogger')

SOURCE_EXTENSIONS = {".c", ".cc", ".cp", ".cpp", ".cxx", ".c++", ".o", ".so",
                     ".a"}

# Replace gcc/g++ build target options with values accepted by Clang.
REPLACE_OPTIONS_MAP = {
    '-mips32': ['-target', 'mips', '-mips32'],
    '-mips64': ['-target', 'mips64', '-mips64'],
    '-mpowerpc': ['-target', 'powerpc'],
    '-mpowerpc64': ['-target', 'powerpc64']
}

# The compilation flags of which the prefix is any of these regular expressions
# will not be included in the output Clang command.
# These flags should be ignored only in case the original compiler is clang.
IGNORED_OPTIONS_CLANG = [
    # Clang gives different warnings than GCC. Thus if these flags are kept,
    # '-Werror', '-pedantic-errors' the analysis with Clang can fail even
    # if the compilation passes with GCC.
    '-Werror',
    '-pedantic-errors',

    # Remove '-w' the option supressing the warnings.
    # This suppressing mechanism is independent of
    # checker enabling/disabling (-W, -W-no), and
    # cannot be overridden by those.
    '-w'
]

# The compilation flags of which the prefix is any of these regular expressions
# will not be included in the output Clang command.
# These flags should be ignored only in case the original compiler is gcc.
IGNORED_OPTIONS_GCC = [
    # --- UNKNOWN BY CLANG --- #
    '-fallow-fetchr-insn',
    '-fcall-saved-',
    '-fcond-mismatch',
    '-fconserve-stack',
    '-fcrossjumping',
    '-fcse-follow-jumps',
    '-fcse-skip-blocks',
    '-fcx-limited-range$',
    '-fext-.*-literals',
    '-ffixed-r2',
    '-ffp$',
    '-mfp16-format',
    '-fgcse-lm',
    '-fhoist-adjacent-loads',
    '-findirect-inlining',
    '-finline-limit',
    '-finline-local-initialisers',
    '-fipa-sra',
    '-fmacro-prefix-map',
    '-fmerge-constants',
    '-fno-aggressive-loop-optimizations',
    '-f(no-)?allow-store-data-races',
    '-fno-canonical-system-headers',
    '-f(no-)?code-hoisting',
    '-fno-delete-null-pointer-checks',
    '-fno-defer-pop',
    '-fno-extended-identifiers',
    '-fno-jump-table',
    '-fno-keep-inline-dllexport'
    '-fno-keep-static-consts',
    '-fno-lifetime-dse',
    '-f(no-)?reorder-functions',
    '-fno-strength-reduce',
    '-fno-toplevel-reorder',
    '-fno-unit-at-a-time',
    '-fno-var-tracking-assignments',
    '-fno-tree-dominator-opts',
    '-fobjc-link-runtime',
    '-fpartial-inlining',
    '-fpeephole2',
    '-fr$',
    '-fregmove',
    '-frename-registers',
    '-frerun-cse-after-loop',
    '-fs$',
    '-fsanitize=bounds-strict',
    '-fsched-pressure',
    '-fsched-spec',
    '-fstack-usage',
    '-fstack-reuse',
    '-fthread-jumps',
    '-ftree-pre',
    '-ftree-switch-conversion',
    '-ftree-tail-merge',
    '-m(no-)?abm',
    '-m(no-)?sdata',
    '-m(no-)?spe',
    '-m(no-)?string$',
    '-m(no-)?dsbt',
    '-m(no-)?fixed-ssp',
    '-m(no-)?pointers-to-nested-functions',
    '-m(no-)?word-relocations',
    '-mno-fp-ret-in-387',
    '-mpreferred-stack-boundary',
    '-mpcrel-func-addr',
    '-mrecord-mcount$',
    '-maccumulate-outgoing-args',
    '-mcall-aixdesc',
    '-mppa3-addr-bug',
    '-mtraceback=',
    '-mtext=',
    '-misa=',
    '-mfunction-return=',
    '-mindirect-branch-register',
    '-mindirect-branch=',
    '-mfix-cortex-m3-ldrd$',
    '-mmultiple$',
    '-msahf$',
    '-mskip-rax-setup$',
    '-mthumb-interwork$',
    '-mupdate$',

    # Deprecated ARM specific option
    # to Generate a stack frame that is compliant
    # with the ARM Procedure Call Standard.
    '-mapcs',
    '-fno-merge-const-bfstores$',
    '-fno-ipa-sra$',
    '-mno-thumb-interwork$',
    # ARM specific option.
    # Prevent the reordering of
    # instructions in the function prologue.
    '-mno-sched-prolog',
    # This is not unknown but we want to preserve asserts to improve the
    # quality of analysis.
    '-DNDEBUG$',

    # --- IGNORED --- #
    '-save-temps',
    # Clang gives different warnings than GCC. Thus if these flags are kept,
    # '-Werror', '-Wno-error', '-pedantic-errors' the analysis with Clang can
    # fail even if the compilation passes with GCC.
    '-Werror',
    '-Wno-error',
    '-pedantic-errors',
    # Profiling flags are different or behave differently in GCC
    '-fprofile',
    # Remove the option disabling the warnings.
    '-w',
    '-g(.+)?$',
    # Link Time Optimization:
    '-flto',
    # MicroBlaze Options:
    '-mxl',
    # PowerPC SPE Options:
    '-mfloat-gprs',
    '-mabi'
]

IGNORED_OPTIONS_GCC = re.compile('|'.join(IGNORED_OPTIONS_GCC))
IGNORED_OPTIONS_CLANG = re.compile('|'.join(IGNORED_OPTIONS_CLANG))

# The compilation flags of which the prefix is any of these regular expressions
# will not be included in the output Clang command. These flags have further
# parameters which are also omitted. The number of parameters is indicated in
# this dictionary.
IGNORED_PARAM_OPTIONS = {
    re.compile('-install_name'): 1,
    re.compile('-exported_symbols_list'): 1,
    re.compile('-current_version'): 1,
    re.compile('-compatibility_version'): 1,
    re.compile('-init$'): 1,
    re.compile('-e$'): 1,
    re.compile('-seg1addr'): 1,
    re.compile('-bundle_loader'): 1,
    re.compile('-multiply_defined'): 1,
    re.compile('-sectorder'): 3,
    re.compile('--param$'): 1,
    re.compile('-u$'): 1,
    re.compile('--serialize-diagnostics'): 1,
    re.compile('-framework'): 1,
    # Darwin linker can be given a file with lists the sources for linking.
    re.compile('-filelist'): 1
}


COMPILE_OPTIONS = [
    '-nostdinc',
    r'-nostdinc\+\+',
    '-pedantic',
    '-O[1-3]',
    '-Os',
    '-std=',
    '-stdlib=',
    '-f',
    '-m',
    '-Wno-',
    '--sysroot=',
    '-sdkroot',
    '--gcc-toolchain=',
    '-pthread'
]

COMPILE_OPTIONS = re.compile('|'.join(COMPILE_OPTIONS))

COMPILE_OPTIONS_MERGED = [
    '--sysroot',
    '-sdkroot',
    '--include',
    '-include',
    '-iquote',
    '-[DIUF]',
    '-idirafter',
    '-isystem',
    '-imacros',
    '-isysroot',
    '-iprefix',
    '-iwithprefix',
    '-iwithprefixbefore'
]

INCLUDE_OPTIONS_MERGED = [
    '-iquote',
    '-[IF]',
    '-isystem',
    '-iprefix',
    '-iwithprefix',
    '-iwithprefixbefore'
]

XCLANG_FLAGS_TO_SKIP = ['-module-file-info',
                        '-S',
                        '-emit-llvm',
                        '-emit-llvm-bc',
                        '-emit-llvm-only',
                        '-emit-llvm-uselists',
                        '-rewrite-objc']

COMPILE_OPTIONS_MERGED = \
    re.compile('(' + '|'.join(COMPILE_OPTIONS_MERGED) + ')')

INCLUDE_OPTIONS_MERGED = \
    re.compile('(' + '|'.join(INCLUDE_OPTIONS_MERGED) + ')')


PRECOMPILATION_OPTION = re.compile('-(E|M[G|T|Q|F|J|P|V|M]*)$')

# Match for all of the compiler flags.
CLANG_OPTIONS = re.compile('.*')


def filter_compiler_includes_extra_args(compiler_flags):
    """Return the list of flags which affect the list of implicit includes.

    compiler_flags -- A list of compiler flags which may affect the list
                      of implicit compiler include paths, like -std=,
                      --sysroot=, -m32, -m64, -nostdinc or -stdlib=.
    """
    # If these options are present in the original build command, they must
    # be forwarded to get_compiler_includes and get_compiler_defines so the
    # resulting includes point to the target that was used in the build.
    pattern = re.compile('-m(32|64)|-std=|-stdlib=|-nostdinc|--driver-mode=')
    extra_opts = list(filter(pattern.match, compiler_flags))

    pos = next((pos for pos, val in enumerate(compiler_flags)
                if val.startswith('--sysroot')), None)
    if pos is not None:
        if compiler_flags[pos] == '--sysroot':
            extra_opts.append('--sysroot=' + compiler_flags[pos + 1])
        else:
            extra_opts.append(compiler_flags[pos])

    return extra_opts


class ImplicitCompilerInfo:
    """
    C/C++ compilers have implicit assumptions about the environment. Especially
    GCC has some built-in options which make build process non-portable to
    other compilers. For example it comes with a set of include paths that are
    implicitly added to all build actions. The list of these paths is also
    configurable by some compiler flags (--sysroot, -x, build target related
    flags, etc.) The goal of this class is to gather and maintain this implicit
    information.
    """

    # Implicit compiler settings (include paths, target triple, etc.) depend on
    # these attributes, so we use them as a dictionary key which maps these
    # attributes to the implicit settings. In the future we may find that some
    # other attributes are also dependencies of implicit compiler info in which
    # case this tuple should be extended.
    ImplicitInfoSpecifierKey = namedtuple(
        'ImplicitInfoSpecifierKey',
        ['compiler', 'language', 'compiler_flags'])

    compiler_info: Dict[ImplicitInfoSpecifierKey, dict] = {}
    compiler_isexecutable = {}
    # Store the already detected compiler version information.
    # If the value is False the compiler is not clang otherwise the value
    # should be a clang version information object.
    compiler_versions = {}

    @staticmethod
    def c():
        return "c"

    @staticmethod
    def cpp():
        return "c++"

    @staticmethod
    def is_executable_compiler(compiler):
        if compiler not in ImplicitCompilerInfo.compiler_isexecutable:
            ImplicitCompilerInfo.compiler_isexecutable[compiler] = \
                find_executable(compiler) is not None

        return ImplicitCompilerInfo.compiler_isexecutable[compiler]

    @staticmethod
    def __get_compiler_err(cmd: List[str]) -> Optional[str]:
        """
        Returns the stderr of a compiler invocation as string
        or None in case of error.
        """
        try:
            proc = subprocess.Popen(
                cmd,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                universal_newlines=True,
                encoding="utf-8",
                errors="ignore")

            # The parameter is usually a compile command in this context which
            # gets a dash ("-") as a compiler flag. This flag makes gcc
            # expecting the source code from the standard input. This is given
            # to communicate() function.
            _, err = proc.communicate("")
            return err
        except OSError as oerr:
            # TODO: shlex.join(cmd) would be more elegant after upgrading to
            # Python 3.8.
            LOG.error(
                "Error during process execution: %s\n%s\n",
                ' '.join(map(shlex.quote, cmd)), oerr.strerror)

    @staticmethod
    def __parse_compiler_includes(compile_cmd: List[str]):
        """
        "gcc -v -E -" prints a set of information about the execution of the
        preprocessor. This output contains the implicitly included paths. This
        function collects and returns these paths from the output of the gcc
        command above. The actual build command is the parameter of this
        function because the list of implicit include paths is affected by some
        compiler flags (e.g. --sysroot, -x, etc.)
        """
        start_mark = "#include <...> search starts here:"
        end_mark = "End of search list."

        include_paths = []
        lines = ImplicitCompilerInfo.__get_compiler_err(compile_cmd)

        if not lines:
            return include_paths

        do_append = False
        for line in lines.splitlines():
            if line.startswith(end_mark):
                break
            if do_append:
                line = line.strip()
                # On OSX there are framework includes,
                # where we need to strip the "(framework directory)" string.
                # For instance:
                # /System/Library/Frameworks (framework directory)
                fpos = line.find("(framework directory)")
                if fpos == -1:
                    include_paths.append(line)
                else:
                    include_paths.append(line[:fpos - 1])

            if line.startswith(start_mark):
                do_append = True

        return include_paths

    @staticmethod
    def get_compiler_includes(compiler, language, compiler_flags):
        """
        Returns a list of default includes of the given compiler.

        compiler -- The compiler binary of which the implicit include paths are
                    fetched.
        language -- The programming language being compiled (e.g. 'c' or 'c++')
        compiler_flags -- the flags used for compilation
        """
        cmd = [compiler, *compiler_flags, '-E', '-x', language, '-', '-v']

        # TODO: shlex.join(cmd) would be more elegant after upgrading to
        # Python 3.8.
        LOG.debug(
            "Retrieving default includes via %s",
            ' '.join(map(shlex.quote, cmd)))
        ICI = ImplicitCompilerInfo
        include_dirs = ICI.__parse_compiler_includes(cmd)

        return list(map(os.path.normpath, include_dirs))

    @staticmethod
    def get_compiler_target(compiler):
        """
        Returns the target triple of the given compiler as a string.

        compiler -- The compiler binary of which the target architecture is
                    fetched.
        """
        lines = ImplicitCompilerInfo.__get_compiler_err([compiler, '-v'])

        if lines is None:
            return ""

        target_label = "Target:"
        target = ""

        for line in lines.splitlines(True):
            line = line.strip().split()
            if len(line) > 1 and line[0] == target_label:
                target = line[1]

        return target

    @staticmethod
    def get_compiler_standard(compiler, language):
        """
        Returns the default compiler standard of the given compiler. The
        standard is determined by the values of __STDC_VERSION__ and
        __cplusplus predefined macros. These values are integers indicating the
        date of the standard. However, GCC supports a GNU extension for each
        standard. For sake of generality we return the GNU extended standard,
        since it should be a superset of the non-extended one, thus applicable
        in a more general manner.

        compiler -- The compiler binary of which the default compiler standard
                    is fetched.
        language -- The programming lenguage being compiled (e.g. 'c' or 'c++')
        """
        VERSION_C = """
#ifdef __STDC_VERSION__
#  if __STDC_VERSION__ >= 201710L
#    error CC_FOUND_STANDARD_VER#17
#  elif __STDC_VERSION__ >= 201112L
#    error CC_FOUND_STANDARD_VER#11
#  elif __STDC_VERSION__ >= 199901L
#    error CC_FOUND_STANDARD_VER#99
#  elif __STDC_VERSION__ >= 199409L
#    error CC_FOUND_STANDARD_VER#94
#  else
#    error CC_FOUND_STANDARD_VER#90
#  endif
#else
#  error CC_FOUND_STANDARD_VER#90
#endif
        """

        VERSION_CPP = """
#ifdef __cplusplus
#  if __cplusplus >= 201703L
#    error CC_FOUND_STANDARD_VER#17
#  elif __cplusplus >= 201402L
#    error CC_FOUND_STANDARD_VER#14
#  elif __cplusplus >= 201103L
#    error CC_FOUND_STANDARD_VER#11
#  elif __cplusplus >= 199711L
#    error CC_FOUND_STANDARD_VER#98
#  else
#    error CC_FOUND_STANDARD_VER#98
#  endif
#else
#  error CC_FOUND_STANDARD_VER#98
#endif
        """

        standard = ""
        with tempfile.NamedTemporaryFile(
                mode='w+',
                suffix=('.c' if language == 'c' else '.cpp'),
                encoding='utf-8') as source:

            with source.file as f:
                f.write(VERSION_C if language == 'c' else VERSION_CPP)

            err = ImplicitCompilerInfo.\
                __get_compiler_err([compiler, source.name])

            if err is not None:
                finding = re.search('CC_FOUND_STANDARD_VER#(.+)', err)
                if finding:
                    standard = finding.group(1)

        if standard:
            if standard == '94':
                # Special case for C94 standard.
                standard = '-std=iso9899:199409'
            else:
                standard = '-std=gnu' \
                    + ('' if language == 'c' else '++') \
                    + standard

        return standard

    @staticmethod
    def dump_compiler_info(file_path: str):
        dumpable = {
            json.dumps(k): v for k, v
            in ImplicitCompilerInfo.compiler_info.items()}

        with open(file_path, 'w', encoding="utf-8", errors="ignore") as f:
            LOG.debug("Writing compiler info into: %s", file_path)
            json.dump(dumpable, f)

    @staticmethod
    def load_compiler_info(file_path: str):
        """Load compiler information from a file."""
        ICI = ImplicitCompilerInfo
        ICI.compiler_info = {}

        contents = load_json(file_path, {})
        for k, v in contents.items():
            k = json.loads(k)
            ICI.compiler_info[
                ICI.ImplicitInfoSpecifierKey(k[0], k[1], tuple(k[2]))] = v

    @staticmethod
    def set(details, compiler_info_file=None):
        """Detect and set the impicit compiler information.

        If compiler_info_file is available the implicit compiler
        information will be loaded and set from it.
        """
        def compiler_info_key(details):
            extra_opts = tuple(sorted(filter_compiler_includes_extra_args(
                details['analyzer_options'])))

            return ICI.ImplicitInfoSpecifierKey(
                details['compiler'], details['lang'], extra_opts)

        ICI = ImplicitCompilerInfo
        iisk = compiler_info_key(details)

        if compiler_info_file and os.path.exists(compiler_info_file):
            # Compiler info file exists, load it.
            ICI.load_compiler_info(compiler_info_file)
        else:
            if iisk not in ICI.compiler_info:
                ICI.compiler_info[iisk] = {
                    'compiler_includes': ICI.get_compiler_includes(
                        iisk.compiler, iisk.language, iisk.compiler_flags),
                    'compiler_standard': ICI.get_compiler_standard(
                        iisk.compiler, iisk.language),
                    'target': ICI.get_compiler_target(iisk.compiler)
                }

        for k, v in ICI.compiler_info.get(iisk, {}).items():
            if not details.get(k):
                details[k] = v

    @staticmethod
    def get():
        return ImplicitCompilerInfo.compiler_info


class OptionIterator:

    def __init__(self, args):
        self._item = None
        self._it = iter(args)

    def __next__(self):
        self._item = next(self._it)
        return self

    next = __next__

    def __iter__(self):
        return self

    @property
    def item(self):
        return self._item


def get_language(extension):
    # TODO: There are even more in the man page of gcc.
    mapping = {'.c': 'c',
               '.cp': 'c++',
               '.cpp': 'c++',
               '.cxx': 'c++',
               '.txx': 'c++',
               '.cc': 'c++',
               '.C': 'c++',
               '.ii': 'c++',
               '.m': 'objective-c',
               '.mm': 'objective-c++'}
    return mapping.get(extension)


def determine_compiler(gcc_command, is_executable_compiler_fun):
    """
    This function determines the compiler from the given compilation command.
    If the first part of the gcc_command is ccache invocation then the rest
    should be a complete compilation command.

    CCache may have three forms:
    1. ccache g++ main.cpp
    2. ccache main.cpp
    3. /usr/lib/ccache/gcc main.cpp
    In the first case this function drops "ccache" from gcc_command and returns
    the next compiler name.
    In the second case the compiler can be given by config files or an
    environment variable. Currently we don't handle this version, and in this
    case the compiler remanis "ccache" and the gcc_command is not changed.
    The two cases are distinguished by checking whether the second parameter is
    an executable or not.
    In the third case gcc is a symlink to ccache, but we can handle
    it as a normal compiler.

    gcc_command -- A split build action as a list which may or may not start
                   with ccache.

    TODO: The second case could be handled if there was a way for querying the
    used compiler from ccache. This can be configured for ccache in config
    files or environment variables.
    """
    if gcc_command[0].endswith('ccache'):
        if is_executable_compiler_fun(gcc_command[1]):
            return gcc_command[1]

    return gcc_command[0]


def __is_not_include_fixed(dirname):
    """
    This function returns True in case the given dirname is NOT a GCC-specific
    include-fixed directory containing standard headers.
    """
    return os.path.basename(os.path.normpath(dirname)) != 'include-fixed'


def __contains_no_intrinsic_headers(dirname):
    """
    Returns True if the given directory doesn't contain any intrinsic headers.
    """
    if not os.path.exists(dirname):
        return True
    if glob.glob(os.path.join(dirname, "*intrin.h")):
        return False
    return True


def __collect_clang_compile_opts(flag_iterator, details):
    """Collect all the options for clang do not filter anything."""
    if CLANG_OPTIONS.match(flag_iterator.item):
        details['analyzer_options'].append(flag_iterator.item)
        return True


def __collect_transform_xclang_opts(flag_iterator, details):
    """Some specific -Xclang constucts need to be filtered out.

       To generate the proper plist reports and not LLVM IR or
       ASCII text as an output the flags need to be removed.
    """
    if flag_iterator.item == "-Xclang":
        next(flag_iterator)
        next_flag = flag_iterator.item
        if next_flag in XCLANG_FLAGS_TO_SKIP:
            return True

        details['analyzer_options'].extend(["-Xclang", next_flag])
        return True

    return False


def __collect_transform_include_opts(flag_iterator, details):
    """
    This function collects the compilation (i.e. not linker or preprocessor)
    flags to the buildaction.
    """

    m = COMPILE_OPTIONS_MERGED.match(flag_iterator.item)

    if not m:
        return False

    flag = m.group(0)
    together = len(flag) != len(flag_iterator.item)

    if together:
        param = flag_iterator.item[len(flag):]
    else:
        next(flag_iterator)
        param = flag_iterator.item

    # The .plist file contains a section with a list of files. For some
    # further actions these need to be given with an absolute path. Clang
    # prints them with absolute path if the original compiler invocation
    # was given absolute paths.
    # TODO: If Clang will be extended with an extra analyzer option in
    # order to print these absolute paths natively, this conversion will
    # not be necessary.
    flags_with_path = ['-I', '-idirafter', '-iquote', '-isysroot', '-isystem',
                       '-sysroot', '--sysroot']
    if flag in flags_with_path and ('sysroot' in flag or param[0] != '='):
        # --sysroot format can be --sysroot=/path/to/include in this case
        # before the normalization the '=' sign must be removed.
        # We put back the original
        # --sysroot=/path/to/include as
        # --sysroot /path/to/include
        # which is a valid format too.
        if param[0] == '=':
            param = param[1:]
            together = False
        param = os.path.normpath(os.path.join(details['directory'], param))

    details['analyzer_options'].extend(
        [flag + param] if together else [flag, param])

    return True


def __collect_compile_opts(flag_iterator, details):
    """
    This function collects the compilation (i.e. not linker or preprocessor)
    flags to the buildaction.
    """
    if COMPILE_OPTIONS.match(flag_iterator.item):
        details['analyzer_options'].append(flag_iterator.item)
        return True

    return False


def __skip_sources(flag_iterator, _):
    """
    This function skips the compiled source file names (i.e. the arguments
    which don't start with a dash character).
    """
    if flag_iterator.item[0] != '-':
        return True

    return False


def __determine_action_type(flag_iterator, details):
    """
    This function determines whether this is a preprocessing, compilation or
    linking action and sets it in the buildaction object. If the action type is
    set to COMPILE earlier then we don't set it to anything else.
    """
    if flag_iterator.item == '-c':
        details['action_type'] = BuildAction.COMPILE
        return True
    elif flag_iterator.item.startswith('-print-prog-name'):
        if details['action_type'] != BuildAction.COMPILE:
            details['action_type'] = BuildAction.INFO
        return True
    elif PRECOMPILATION_OPTION.match(flag_iterator.item):
        if details['action_type'] != BuildAction.COMPILE:
            details['action_type'] = BuildAction.PREPROCESS
        return True

    return False


def __get_arch(flag_iterator, details):
    """
    This function consumes -arch flag which is followed by the target
    architecture. This is then collected to the buildaction object.
    """
    # TODO: Is this really a target architecture? Have we seen this flag being
    # used in a real project? This -arch flag is not really documented among
    # GCC flags.
    # Where do we use this architecture during analysis and why?
    if flag_iterator.item == '-arch':
        next(flag_iterator)
        details['arch'] = flag_iterator.item
        return True

    return False


def __get_target(flag_iterator, details):
    """
    This function consumes --target or -target flag which is followed by the
    compilation target architecture.
    This target might be different from the default compilation target
    collected from the compiler if cross compilation is done for
    another target.
    This is then collected to the buildaction object.
    """
    if flag_iterator.item in ['--target', '-target']:
        next(flag_iterator)
        details['compilation_target'] = flag_iterator.item
        return True

    return False


def __get_language(flag_iterator, details):
    """
    This function consumes -x flag which is followed by the language. This
    language is then collected to the buildaction object.
    """
    # TODO: Known issue: a -x flag may precede all source files in the build
    # command with different languages.
    if flag_iterator.item.startswith('-x'):
        if flag_iterator.item == '-x':
            next(flag_iterator)
            details['lang'] = flag_iterator.item
        else:
            details['lang'] = flag_iterator.item[2:]  # 2 == len('-x')
        return True
    return False


def __get_output(flag_iterator, details):
    """
    This function consumes -o flag which is followed by the output file of the
    action. This file is then collected to the buildaction object.
    """
    if flag_iterator.item == '-o':
        next(flag_iterator)
        details['output'] = flag_iterator.item
        return True

    return False


def __replace(flag_iterator, details):
    """
    This function extends the analyzer options list with the corresponding
    replacement based on REPLACE_OPTIONS_MAP if the flag_iterator is currently
    pointing to a flag to replace.
    """
    value = REPLACE_OPTIONS_MAP.get(flag_iterator.item)

    if value:
        details['analyzer_options'].extend(value)

    return bool(value)


def __skip_clang(flag_iterator, _):
    """
    This function skips the flag pointed by the given flag_iterator with its
    parameters if any.
    """
    if IGNORED_OPTIONS_CLANG.match(flag_iterator.item):
        return True

    return False


def __skip_gcc(flag_iterator, _):
    """
    This function skips the flag pointed by the given flag_iterator with its
    parameters if any.
    """
    if IGNORED_OPTIONS_GCC.match(flag_iterator.item):
        return True

    for pattern, arg_num in IGNORED_PARAM_OPTIONS.items():
        if pattern.match(flag_iterator.item):
            for _ in range(arg_num):
                next(flag_iterator)
            return True

    return False


def parse_options(compilation_db_entry,
                  compiler_info_file=None,
                  keep_gcc_include_fixed=False,
                  keep_gcc_intrin=False,
                  get_clangsa_version_func=None,
                  analyzer_clang_version=None):
    """
    This function parses a GCC compilation action and returns a BuildAction
    object which can be the input of Clang analyzer tools.

    compilation_db_entry -- An entry from a valid compilation database JSON
                            file, i.e. a dictionary with the compilation
                            command, the compiled file and the current working
                            directory.
    compiler_info_file -- Contains the path to a compiler info file.
    keep_gcc_include_fixed -- There are some implicit include paths which are
                              only used by GCC (include-fixed). This flag
                              determines whether these should be kept among
                              the implicit include paths.
    keep_gcc_intrin -- There are some implicit include paths which contain
                       GCC-specific header files (those which end with
                       intrin.h). This flag determines whether these should be
                       kept among the implicit include paths. Use this flag if
                       Clang analysis fails with error message related to
                       __builtin symbols.
    get_clangsa_version_func -- Is a function which should return the
                            version information for a clang compiler.
                            It requires the compiler binary and an env.
                            get_clangsa_version_func(compiler_binary, env)
                            Should return false for a non clang compiler.
    analyzer_clang_version -- version information about the clang which is
                              used to execute the analysis
    """
    details = {
        'analyzer_options': [],
        'compiler_includes': [],
        'compiler_standard': '',
        'compilation_target': '',  # Compilation target in the compilation cmd.
        'analyzer_type': -1,
        'original_command': '',
        'directory': '',
        'output': '',
        'lang': None,
        'arch': '',  # Target in the compile command set by -arch.
        'target': '',
        'source': ''}

    if 'arguments' in compilation_db_entry:
        gcc_command = compilation_db_entry['arguments']
        details['original_command'] = \
            ' '.join([shlex.quote(x) for x in gcc_command])
    elif 'command' in compilation_db_entry:
        details['original_command'] = compilation_db_entry['command']
        gcc_command = shlex.split(compilation_db_entry['command'])
    else:
        raise KeyError("No valid 'command' or 'arguments' entry found!")

    details['directory'] = compilation_db_entry['directory']
    details['action_type'] = None
    details['compiler'] =\
        determine_compiler(gcc_command,
                           ImplicitCompilerInfo.is_executable_compiler)
    if '++' in os.path.basename(details['compiler']):
        details['lang'] = 'c++'

    # Source files are skipped first so they are not collected
    # with the other compiler flags together. Source file is handled
    # separately from the compile command json.
    clang_flag_collectors = [
        __skip_sources,
        __skip_clang,
        __collect_transform_xclang_opts,
        __get_output,
        __determine_action_type,
        __get_arch,
        __get_target,
        __get_language,
        __collect_transform_include_opts,
        __collect_clang_compile_opts
        ]

    gcc_flag_transformers = [
        __skip_gcc,
        __replace,
        __collect_transform_include_opts,
        __collect_compile_opts,
        __determine_action_type,
        __skip_sources,
        __get_arch,
        __get_target,
        __get_language,
        __get_output]

    flag_processors = gcc_flag_transformers

    compiler_version_info = \
        ImplicitCompilerInfo.compiler_versions.get(
            details['compiler'], False)

    if not compiler_version_info and get_clangsa_version_func:

        # did not find in the cache yet
        try:
            compiler_version_info = \
                get_clangsa_version_func(details['compiler'])
        except (subprocess.CalledProcessError, OSError) as cerr:
            LOG.error('Failed to get and parse version of: %s',
                      details['compiler'])
            LOG.error(cerr)
            compiler_version_info = False

    ImplicitCompilerInfo.compiler_versions[details['compiler']] \
        = compiler_version_info

    using_same_clang_to_compile_and_analyze = False
    compiler_clang = \
        ImplicitCompilerInfo.compiler_versions.get(details['compiler'])

    if compiler_clang:
        # Based on the version information the compiler is clang.
        flag_processors = clang_flag_collectors

        if analyzer_clang_version and \
            compiler_clang.installed_dir == \
                analyzer_clang_version.installed_dir:
            LOG.debug("Same clang is used for compilation and analysis.")
            using_same_clang_to_compile_and_analyze = True

    for it in OptionIterator(gcc_command[1:]):
        for flag_processor in flag_processors:
            if flag_processor(it, details):
                break
        else:
            pass
            # print('Unhandled argument: ' + it.item)

    if details['action_type'] is None:
        details['action_type'] = BuildAction.COMPILE

    details['source'] = compilation_db_entry['file']
    # Calculate absolute path to file if not already abs
    if not Path(details['source']).is_absolute():
        details['source'] = str(Path(
            Path(compilation_db_entry['directory']).absolute(),
            compilation_db_entry['file']).resolve())

    # In case the file attribute in the entry is empty.
    if details['source'] == '.':
        details['source'] = ''

    lang = get_language(os.path.splitext(details['source'])[1])
    if lang:
        if details['lang'] is None:
            details['lang'] = lang
    else:
        details['action_type'] = BuildAction.LINK

    # Option parser detects target architecture but does not know about the
    # language during parsing. Set the collected compilation target for the
    # language detected language.
    details['target'] = details['compilation_target']

    # With gcc-toolchain a non default compiler toolchain can be set. Clang
    # will search for include paths and libraries based on the gcc-toolchain
    # parameter. Detecting extra include paths from the host compiler could
    # conflict with this.

    # For example if the compiler in the compile command is clang and
    # gcc-toolchain is set we will get the include paths for clang and not for
    # the compiler set in gcc-toolchain. This can cause missing headers during
    # the analysis.

    toolchain = \
        gcc_toolchain.toolchain_in_args(details['analyzer_options'])

    # Store the compiler built in include paths and defines.
    # If clang compiler is used for compilation and analysis, or language is
    # not recognized, do not collect the implicit include paths.
    if ((not toolchain and not using_same_clang_to_compile_and_analyze) or
            (compiler_info_file and os.path.exists(compiler_info_file))) and \
            details['lang']:
        ImplicitCompilerInfo.set(details, compiler_info_file)

    if not keep_gcc_include_fixed:
        details['compiler_includes'] = list(filter(
            __is_not_include_fixed,
            details['compiler_includes']))

    if not keep_gcc_intrin:
        details['compiler_includes'] = list(filter(
            __contains_no_intrinsic_headers,
            details['compiler_includes']))

        # filter out intrin directories
        aop_without_intrin = []
        analyzer_options = iter(details['analyzer_options'])

        for aopt in analyzer_options:
            m = INCLUDE_OPTIONS_MERGED.match(aopt)
            if m:
                flag = m.group(0)
                together = len(flag) != len(aopt)

                if together:
                    value = aopt[len(flag):]
                else:
                    flag = aopt
                    value = next(analyzer_options)
                if os.path.isdir(value) and __contains_no_intrinsic_headers(
                        value) or not os.path.isdir(value):
                    if together:
                        aop_without_intrin.append(aopt)
                    else:
                        aop_without_intrin.append(flag)
                        aop_without_intrin.append(value)
            else:
                # no match
                aop_without_intrin.append(aopt)

        details['analyzer_options'] = aop_without_intrin

    return BuildAction(**details)


def process_response_file(response_file):
    """
    Return list of options and source files from the given response file.
    """
    with open(response_file, encoding="utf-8", errors="ignore") as r_file:
        options = shlex.split(r_file)

    sources = [opt for opt in options if not opt.startswith('-') and
               os.path.splitext(opt)[1].lower() in SOURCE_EXTENSIONS]

    return options, sources


def extend_compilation_database_entries(compilation_database):
    """
    Loop through the compilation database entries and whether compilation
    command contains a response file we read those files and replace the
    response file with the options from the file.
    """
    entries = []
    for entry in compilation_database:
        if 'command' in entry and '@' in entry['command']:
            cmd = []
            source_files = []
            source_dir = entry['directory']

            options = shlex.split(entry['command'])
            for opt in options:
                if opt.startswith('@'):
                    response_file = os.path.join(source_dir, opt[1:])
                    if not os.path.exists(response_file):
                        LOG.warning("Response file '%s' does not exists.",
                                    response_file)
                        continue

                    opts, sources = process_response_file(response_file)
                    cmd.extend([shlex.quote(x) for x in opts])
                    source_files.extend(sources)
                else:
                    cmd.append(shlex.quote(opt))

            entry['command'] = ' '.join(cmd)

            if entry['file'].startswith('@'):
                for source_file in source_files:
                    new_entry = dict(entry)
                    new_entry['file'] = source_file
                    entries.append(new_entry)
                continue

        entries.append(entry)

    return entries


class CompileCommandEncoder(json.JSONEncoder):
    """JSON serializer for objects not serializable by default json code"""
    # pylint: disable=method-hidden
    def default(self, o):
        if isinstance(o, BuildAction):
            return o.to_dict()
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, o)


class CompileActionUniqueingType(Enum):
    NONE = 0  # Full Action text
    SOURCE_ALPHA = 1  # Based on source file, uniqueing by
    # alphabetically first output object file name
    SOURCE_REGEX = 2  # Based on source file, uniqueing by regex filter
    STRICT = 3  # Gives error in case of duplicate
    SYMLINK = 4  # Based on source file, uniqueing by
    # recognizing symlink and remove duplication


def parse_unique_log(compilation_database,
                     report_dir,
                     compile_uniqueing="none",
                     compiler_info_file=None,
                     keep_gcc_include_fixed=False,
                     keep_gcc_intrin=False,
                     analysis_skip_handlers=None,
                     pre_analysis_skip_handlers=None,
                     ctu_or_stats_enabled=False,
                     analyzer_clang_version=None):
    """
    This function reads up the compilation_database
    and returns with a list of build actions that is
    prepared (uniqued and skipped) for clang execution together
    with the number of skipped compile commands.

    That means that gcc specific parameters are filtered out
    and gcc built in targets and include paths are added.
    It also filters out duplicate compilation actions based on the
    compile_uniqueing parameter.
    This function also dumps auto-detected the compiler info
    into <report_dir>/compiler_info.json.

    compilation_database -- A compilation database as a list of dict objects.
                            These object should contain "file", "dictionary"
                            and "command" keys. The "command" may be replaced
                            by "arguments" which is a split command. Older
                            versions of intercept-build provide the build
                            command this way.
    report_dir  -- The output report directory. The compiler infos
                   will be written to <report_dir>/compiler.info.json.
    compile_uniqueing -- Compilation database uniqueing mode.
                         If there are more than one compile commands for a
                         target file, only a single one is kept.
    compiler_info_file -- compiler_info.json. If exists, it will be used for
                    analysis.
    keep_gcc_include_fixed -- There are some implicit include paths which are
                              only used by GCC (include-fixed). This flag
                              determines whether these should be kept among
                              the implicit include paths.
    keep_gcc_intrin -- There are some implicit include paths which contain
                       GCC-specific header files (those which end with
                       intrin.h). This flag determines whether these should be
                       kept among the implicit include paths. Use this flag if
                       Clang analysis fails with error message related to
                       __builtin symbols.

    Separate skip handlers are required because it is possible that different
    files are skipped during pre analysis and the actual analysis. In the
    pre analysis step nothing should be skipped to collect the required
    information for the analysis step where not all the files are analyzed.

    analysis_skip_handlers -- skip handlers for files which should be skipped
                             during analysis
    pre_analysis_skip_handlers -- skip handlers for files wich should be
                                 skipped during pre analysis
    ctu_or_stats_enabled -- ctu or statistics based analysis was enabled
                            influences the behavior which files are skipped.
    analyzer_clang_version -- version information about the clang which is
                              used to execute the analysis
    """
    try:
        uniqued_build_actions = dict()

        if compile_uniqueing == "alpha":
            build_action_uniqueing = CompileActionUniqueingType.SOURCE_ALPHA
        elif compile_uniqueing == "none":
            build_action_uniqueing = CompileActionUniqueingType.NONE
        elif compile_uniqueing == "strict":
            build_action_uniqueing = CompileActionUniqueingType.STRICT
        elif compile_uniqueing == "symlink":
            build_action_uniqueing = CompileActionUniqueingType.SYMLINK
        else:
            build_action_uniqueing = CompileActionUniqueingType.SOURCE_REGEX
            uniqueing_re = re.compile(compile_uniqueing)

        skipped_cmp_cmd_count = 0

        for entry in extend_compilation_database_entries(compilation_database):
            # Parse compilation db entry into an action object
            action = parse_options(entry,
                                   compiler_info_file,
                                   keep_gcc_include_fixed,
                                   keep_gcc_intrin,
                                   clangsa.version.get,
                                   analyzer_clang_version)

            # Skip parsing the compilaton commands if it should be skipped
            # at both analysis phases (pre analysis and analysis).
            # Skipping of the compile commands is done differently if no
            # CTU or statistics related feature was enabled.
            if analysis_skip_handlers \
                and analysis_skip_handlers.should_skip(action.source) \
                and (not ctu_or_stats_enabled or pre_analysis_skip_handlers
                     and pre_analysis_skip_handlers.should_skip(
                         action.source)):
                skipped_cmp_cmd_count += 1
                LOG.debug("skipping:", action.source)
                continue

            if not action.lang:
                skipped_cmp_cmd_count += 1
                continue
            if action.action_type != BuildAction.COMPILE:
                skipped_cmp_cmd_count += 1
                continue
            if build_action_uniqueing == CompileActionUniqueingType.NONE:
                if action not in uniqued_build_actions:
                    uniqued_build_actions[action] = action
            elif build_action_uniqueing == CompileActionUniqueingType.STRICT:
                if action.source not in uniqued_build_actions:
                    uniqued_build_actions[action.source] = action
                else:
                    LOG.error("Build Action uniqueing failed"
                              " as both '%s' and '%s'",
                              uniqued_build_actions[action.source]
                              .original_command,
                              action.original_command)
                    sys.exit(1)
            elif build_action_uniqueing ==\
                    CompileActionUniqueingType.SOURCE_ALPHA:
                if action.source not in uniqued_build_actions:
                    uniqued_build_actions[action.source] = action
                elif action.output <\
                        uniqued_build_actions[action.source].output:
                    uniqued_build_actions[action.source] = action
            elif build_action_uniqueing == CompileActionUniqueingType.SYMLINK:
                real_path = os.path.realpath(entry['file'])
                if real_path not in uniqued_build_actions:
                    uniqued_build_actions[real_path] = action
            elif build_action_uniqueing ==\
                    CompileActionUniqueingType.SOURCE_REGEX:
                LOG.debug("uniqueing regex")
                if action.source not in uniqued_build_actions:
                    uniqued_build_actions[action.source] = action
                elif uniqueing_re.match(action.original_command) and\
                    not uniqueing_re.match(
                        uniqued_build_actions[action.source].original_command):
                    uniqued_build_actions[action.source] = action
                elif uniqueing_re.match(action.original_command) and\
                    uniqueing_re.match(
                        uniqued_build_actions[action.source].original_command):
                    LOG.error("Build Action uniqueing failed as both \n %s"
                              "\n and \n %s \n match regex pattern:%s",
                              uniqued_build_actions[action.source].
                              original_command,
                              action.original_command,
                              compile_uniqueing)
                    sys.exit(1)

        ImplicitCompilerInfo.dump_compiler_info(
            os.path.join(report_dir, "compiler_info.json"))

        LOG.debug('Parsing log file done.')
        return list(uniqued_build_actions.values()), skipped_cmp_cmd_count

    except (ValueError, KeyError, TypeError) as ex:
        if not compilation_database:
            LOG.error('The compile database is empty.')
        else:
            LOG.error('The compile database is not valid.')
        LOG.debug(traceback.format_exc())
        LOG.debug(ex)
        sys.exit(1)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

"""
This module defines the list of permissions of a CodeChecker server, and
provides the handling of permission matching and databases.
"""


from abc import ABCMeta
from abc import abstractmethod

from sqlalchemy import and_, func

from codechecker_api_shared.ttypes import Permission as PermissionEnum

from codechecker_common.logger import get_logger

LOG = get_logger('server')
config_db_model = None  # Module will be loaded later...


class Permission(metaclass=ABCMeta):
    """
    The base class for a permission declaration.
    """

    def __init__(self, name, default_enable=True,
                 inherited_from=None, managed_by=None, manages_self=False):
        """
        Creates the definition of a new permission.

        :param name:           The name of the permission
        :param default_enable: If False, only the people explicitly given this
          permission are marked as having it. If True, an empty list of people
          given the permission means that everyone has the permission, if the
          server is running in authentication ENABLED mode.
        :param inherited_from: The list of permissions which automatically
          imply this permission. (Disjunctive list, i.e. if the user has
          either of these permissions specified, they have the current one
          too.)
          Permissions can only be inherited from another permission in the
          same, or a broader scope.
        :param managed_by:     The list of permissions from which the user must
          have at least one to manage other users' authorisation on this
          permission.
          Permissions can only be managed by having permissions in the same,
          or a broader scope.
        :param manages_self:   If True, people who have this permission also
          can manage this permission.
        """
        self.name = name
        self.default_enable = default_enable

        if inherited_from is None:
            inherited_from = []
        if managed_by is None:
            managed_by = []

        # Permission can be managed by itself.
        if manages_self:
            managed_by = [self] + managed_by

        # The SUPERUSER permission is a hardcoded special case.
        if name == 'SUPERUSER':
            # SUPERUSER is the maximum rank, it cannot be inherited.
            self.inherited_from = []
        else:
            # Granting SUPERUSER automatically grants every other permission.
            if SUPERUSER not in inherited_from:
                inherited_from = [SUPERUSER] + inherited_from

            # SUPERUSERs can manage every permission by default.
            if SUPERUSER not in managed_by:
                managed_by = [SUPERUSER] + managed_by

            self.inherited_from = inherited_from

        self.managed_by = managed_by

    # Contains the list of arguments required by __call__ to initialise a
    # permission handler.
    # (Unfortunately, inspect.getargspec() doesn't seem to work when
    # called from the API handler.)
    CALL_ARGS = []

    @abstractmethod
    def __call__(self, *args, **kwargs):
        """
        Instantiate a PermissionHandler class for the current permission.
        This instantiated object can be used to query and manage permissions.

        This method is overridden in subclasses so that permissions of
        different scopes can create an appropriate handler.

        The amount of arguments for this function may vary, depending on the
        scope of the permission.
        """
        pass


class PermissionHandler(metaclass=ABCMeta):
    """
    The abstract base class for an object that interfaces with the permission
    database to query and manage permissions.
    """

    def __init__(self, permission):
        """
        Create the Permission Handler.

        :param permission:  The Permission object that instantiated this
          handler.
        """
        self._permission = permission
        self._perm_name = permission.name

        # The actual database managing methods (beginning with __) need the
        # database model to be available, but we cannot say
        # "import config_db_model" at the top of this file because the database
        # module needs this module to be fully loaded (permissions constants
        # defined) to set up the schema.
        #
        # Thus, a deferred loading is used here, the first time a permission
        # handler is initialized. Most likely this import does nothing,
        # as the server already executed loading the config_db_model,
        # so we just set the name to properly point to the module object.
        global config_db_model
        if config_db_model is None:
            LOG.debug("Handler initiated for first time, loading ORM...")
            from .database import config_db_model as ConfigDB
            config_db_model = ConfigDB

    # These high-level methods are used by client code. These contain
    # control flow that are shared for every permission handler
    # (such as handling default_enabled), but these don't access any
    # database backend.
    def add_permission(self, auth_name, is_group=False, user_name='Anonymous'):
        """
        Registers the current permission for the given authentication
        identifier, which is either a username, or a group name, depending on
        is_group.
        """
        if auth_name == "*":
            raise ValueError("'*' is a special token which can NOT be "
                             "directly added to the permission table.")

        added = self._add_perm_impl(auth_name, is_group)

        if added:
            LOG.info("Permission '%s' added for %s '%s' by '%s'.",
                     self._perm_name, 'group' if is_group else 'user',
                     auth_name, user_name)

            # Set the invariant for default_enable.
            if self._permission.default_enable:
                users, groups = self.list_permitted()

                # If a user was added (and this is the first add to a default
                # enabled permission) the table must contain '*' and the user,
                # otherwise a '*' and the currently added group.
                if (not is_group and len(users) == 2 and not groups) or \
                        (is_group and len(users) == 1 and len(groups) == 1):
                    self._rem_perm_impl("*", False)
        else:
            LOG.info("Permission '%s' already added for %s '%s'!",
                     self._perm_name, 'group' if is_group else 'user',
                     auth_name)

    def remove_permission(self, auth_name, is_group=False,
                          user_name='Anonymous'):
        """
        Removes the current permission from the given authentication
        identifier.
        """
        if auth_name == "*":
            raise ValueError("'*' is a special token which can NOT be "
                             "directly removed from the permission table.")

        removed = self._rem_perm_impl(auth_name, is_group)

        if removed:
            LOG.info("Permission '%s' removed from %s '%s' by '%s'.",
                     self._perm_name, 'group' if is_group else 'user',
                     auth_name, user_name)

            # Set the invariant for default_enable.
            if self._permission.default_enable:
                users, groups = self.list_permitted()
                if not users and not groups:
                    self._add_perm_impl("*", False)
        else:
            LOG.info("Permission '%s' already removed from %s '%s'!",
                     self._perm_name, 'group' if is_group else 'user',
                     auth_name)

    def has_permission(self, auth_session):
        """
        Returns whether or not the given authenticated user session
        (or None, if authentication is disabled on the server!) is given
        the current permission.
        """
        if not auth_session:
            # If the user does not have an auth_session it means it is a guest
            # and the server is running in authentication disabled mode.
            # All permissions are automatically granted in this case.
            return True
        elif auth_session.is_root and self._perm_name == 'SUPERUSER':
            # The special master superuser (root) automatically has the
            # SUPERUSER permission.
            return True

        name = self._has_perm_impl([auth_session.user], False)
        groups = self._has_perm_impl(auth_session.groups, True)

        if not name and not groups and self._permission.default_enable:
            # Default enabled permission work in a way that if no one has the
            # permission, everyone has it.
            # ("No-one has the permission" is represented as a * user having
            # the permission, this invariant kept up by add() and remove().)
            return self._has_perm_impl(["*"], False)

        return name or groups

    def list_permitted(self):
        """
        Returns a pair of usernames and groups that are given the current
        permission.
        """
        records = self._list_authorised_impl()
        users = []
        groups = []
        for name, is_group in records:
            if is_group:
                groups.append(name)
            else:
                users.append(name)

        return users, groups

    # These abstract methods are overridden in the subclasses to interface
    # with some database backend in a particular way, depending on the
    # permission's scope.
    @abstractmethod
    def _add_perm_impl(self, auth_name, is_group=False):
        pass

    @abstractmethod
    def _rem_perm_impl(self, auth_name, is_group=False):
        pass

    @abstractmethod
    def _has_perm_impl(self, auth_names, are_groups=False):
        pass

    @abstractmethod
    def _list_authorised_impl(self):
        pass


class SystemPermission(Permission):
    """
    Represents a permission which scope is the entire configuration database,
    essentially the full server.
    """

    class Handler(PermissionHandler):
        """
        This class is used to query server-wide permissions.
        """

        def __init__(self, permission, config_db_session):
            """
            Create a new system-wide permission handler.

            :param permission:        The Permission object that instantiated
              this handler.
            :param config_db_session: A database session that refers to the
              configuration database of the server.
            """
            super(SystemPermission.Handler, self).__init__(permission)
            self.__session = config_db_session

        def __get_perm_record(self, auth_name, is_group):
            SysPerm = config_db_model.SystemPermission

            record = self.__session. \
                query(SysPerm). \
                filter(and_(
                    SysPerm.permission == self._perm_name,
                    func.lower(SysPerm.name) == auth_name.lower(),
                    SysPerm.is_group == is_group
                )).one_or_none()

            return record

        def __get_stored_auth_name_and_permissions(self, auth_name, is_group):
            """
            Query user or group system permission set.

            :param auth_name:   User's name or name of group name case
                                insensitive pattern.
            :param is_group:    Determines that the previous name either a
                                user's name or a group name.
            :returns:           A touple in (name, permission_set) structure.
            """

            SysPerm = config_db_model.SystemPermission

            stored_auth_name = auth_name
            permissions = set()
            for name, permission in self.__session.query(
                    SysPerm.name, SysPerm.permission).filter(and_(
                    func.lower(SysPerm.name) == auth_name.lower(),
                    SysPerm.is_group == is_group)):
                stored_auth_name = name
                permissions.add(permission)

            return (stored_auth_name, permissions)

        def _add_perm_impl(self, auth_name, is_group=False):
            if not auth_name:
                return False

            stored_auth_name, permissions = \
                self.__get_stored_auth_name_and_permissions(
                    auth_name, is_group)

            if not permissions:  # This account have not got permission yet.
                new_permission_record = config_db_model.SystemPermission(
                    self._permission.name, auth_name, is_group)
            else:  # There are at least one permission of the user.
                if self._permission.name in permissions:
                    return False  # Required permission already granted

                new_permission_record = config_db_model.SystemPermission(
                    self._permission.name, stored_auth_name, is_group)

            self.__session.add(new_permission_record)
            return True

        def _rem_perm_impl(self, auth_name, is_group=False):
            if not auth_name:
                return False

            perm_record = self.__get_perm_record(auth_name, is_group)
            if perm_record:
                self.__session.delete(perm_record)
                return True

        def _has_perm_impl(self, auth_names, are_groups=False):
            if not auth_names:
                return False

            auth_names_lower = [name.lower() for name in auth_names]
            SysPerm = config_db_model.SystemPermission
            query = self.__session. \
                query(SysPerm). \
                filter(and_(
                    SysPerm.permission == self._perm_name,
                    func.lower(SysPerm.name).in_(auth_names_lower),
                    SysPerm.is_group == are_groups
                ))

            exists = self.__session.query(query.exists()).scalar()
            return exists

        def _list_authorised_impl(self):
            SysPerm = config_db_model.SystemPermission

            result = self.__session. \
                query(SysPerm). \
                filter(SysPerm.permission == self._perm_name).all()

            return [(p.name, p.is_group) for p in result]

    def __init__(self, name, **kwargs):
        super(SystemPermission, self).__init__(name, **kwargs)

    CALL_ARGS = ['config_db_session']

    def __call__(self, config_db_session):
        """
        Create the permission handler for this system-wide permission,
        for the given configuration database session.
        """

        return SystemPermission.Handler(self, config_db_session)


class ProductPermission(Permission):
    """
    Represents a permission which scope is a particular product.
    """

    class Handler(PermissionHandler):
        """
        This class is used to query product-wide permissions.
        """

        def __init__(self, permission, config_db_session, product_id):
            """
            Create a new product-wide permission handler.

            :param permission:        The Permission object that instantiated
              this handler.
            :param config_db_session: A database session that refers to the
              configuration database of the server.
            :param product_id:        The ID of the product for which the
              permission is instantiated.
            """
            super(ProductPermission.Handler, self).__init__(permission)
            self.__session = config_db_session
            self.__product_id = product_id

        def __get_perm_record(self, auth_name, is_group):
            ProdPerm = config_db_model.ProductPermission

            record = self.__session. \
                query(ProdPerm). \
                filter(and_(
                    ProdPerm.permission == self._perm_name,
                    ProdPerm.product_id == self.__product_id,
                    func.lower(ProdPerm.name) == auth_name.lower(),
                    ProdPerm.is_group == is_group
                )).one_or_none()

            return record

        def __get_stored_auth_name_and_permissions(self, auth_name, is_group):
            """
            Query user or group product permission set.

            :param auth_name:   User's name or name of group name case
                                insensitive pattern.
            :param is_group:    Determines that the previous name either a
                                user's name or a group name.
            :returns:           A touple in (name, permission_set) structure.
            """
            ProdPerm = config_db_model.ProductPermission

            stored_auth_name = auth_name
            permissions = set()
            for name, permission in self.__session.query(
                    ProdPerm.name, ProdPerm.permission).filter(and_(
                    ProdPerm.product_id == self.__product_id,
                    func.lower(ProdPerm.name) == auth_name.lower(),
                    ProdPerm.is_group == is_group)):
                stored_auth_name = name
                permissions.add(permission)

            return (stored_auth_name, permissions)

        def _add_perm_impl(self, auth_name, is_group=False):
            if not auth_name:
                return False

            stored_auth_name, permission_set = \
                self.__get_stored_auth_name_and_permissions(
                    auth_name, is_group)

            if not permission_set:  # This account have not got permission yet.
                new_permission_record = config_db_model.ProductPermission(
                    self._permission.name, self.__product_id, auth_name,
                    is_group)
            else:  # There are at least one permission of the user.
                if self._permission.name in permission_set:
                    return False  # Required permission already granted

                new_permission_record = config_db_model.ProductPermission(
                    self._permission.name, self.__product_id,
                    stored_auth_name, is_group)

            self.__session.add(new_permission_record)
            return True

        def _rem_perm_impl(self, auth_name, is_group=False):
            if not auth_name:
                return False

            perm_record = self.__get_perm_record(auth_name, is_group)
            if perm_record:
                self.__session.delete(perm_record)
                return True

        def _has_perm_impl(self, auth_names, are_groups=False):
            if not auth_names:
                return False

            # Compare authorization names in a case insensitive way.
            auth_names_lower = [name.lower() for name in auth_names]
            ProdPerm = config_db_model.ProductPermission
            query = self.__session. \
                query(ProdPerm). \
                filter(and_(
                    ProdPerm.permission == self._perm_name,
                    ProdPerm.product_id == self.__product_id,
                    func.lower(ProdPerm.name).in_(auth_names_lower),
                    ProdPerm.is_group.is_(are_groups)
                ))

            exists = self.__session.query(query.exists()).scalar()
            return exists

        def _list_authorised_impl(self):
            ProdPerm = config_db_model.ProductPermission

            result = self.__session. \
                query(ProdPerm). \
                filter(and_(
                    ProdPerm.permission == self._perm_name,
                    ProdPerm.product_id == self.__product_id
                )).all()

            return [(p.name, p.is_group) for p in result]

    def __init__(self, name, **kwargs):
        super(ProductPermission, self).__init__(name, **kwargs)

    CALL_ARGS = ['config_db_session', 'productID']

    def __call__(self, config_db_session, productID):
        """
        Create the permission handler for this product-wide permission
        on a particular product, managed through the given configuration
        database session.
        """

        return ProductPermission.Handler(self, config_db_session, productID)

# ---------------------------------------------------------------------------


PERMISSION_SCOPES = {
    'SYSTEM': SystemPermission,
    'PRODUCT': ProductPermission
}

__PERM_TO_API = {}
__API_TO_PERM = {}


def _create_permission(clazz, name, **kwargs):
    """
    Create a new permission and register it.

    :param clazz:  The subclass of Permission which will be instantiated.
    :param name:   The name of the permission. This is corresponding to the
      enum value over the API, usually in full capitals.
    :param kwargs: The rest of the argument dict will be passed as-is to the
      constructor of clazz.
    :return: The created permission instance.
    """

    permission = clazz(name, **kwargs)
    enum_value = PermissionEnum._NAMES_TO_VALUES[name]

    __PERM_TO_API[permission] = enum_value
    __API_TO_PERM[enum_value] = permission

    return permission


def get_permissions(scope=None):
    """
    Returns a list of permissions.

    :param scope: One of the Permission class strings (e.g. 'SYSTEM'), which
      if given, filters the returned list of permissions to only definitions
      of the given scope.
    """

    if scope is not None and scope not in PERMISSION_SCOPES:
        return []

    ret = []
    for _, perm in sorted(__API_TO_PERM.items()):
        if scope is not None and \
                not isinstance(perm, PERMISSION_SCOPES[scope]):
            continue
        ret.append(perm)
    return ret


def permission_from_api_enum(key):
    """
    Returns the Permission object for the associated API enum value.
    """
    return __API_TO_PERM[key]


def api_enum_for_permission(permission):
    """
    Return the API enum value for the given permission object.
    """
    return __PERM_TO_API[permission]


def handler_from_scope_params(permission, extra_params):
    """
    Construct the given permission's handler using the scope-specific extra
    arguments (received from the API). This method filters the extra_params
    to the values actually needed by the handler constructor.
    """

    # Filter the list of arguments of the handler factory __call__()
    # with the arguments given by the API.
    args_to_pass = set(extra_params).intersection(permission.CALL_ARGS)

    kwargs = {key: extra_params[key] for key in args_to_pass}
    return permission(**kwargs)


def initialise_defaults(scope, extra_params):
    """
    Helper function which creates the default-permission records in the
    database for a given scope and for the scope-specific extra params.

    This is usually used at server start (for SYSTEM permissions) and the
    creation of a new product (for PRODUCT permissions), etc.
    """

    perms = [perm for perm in get_permissions(scope)]

    for perm in perms:
        handler = handler_from_scope_params(perm, extra_params)
        users, groups = handler.list_permitted()

        if perm.default_enable and not users and not groups:
            # Calling the implementation method directly as this call is
            # needed to set an invariant which the interface method would
            # presume to be already standing.
            handler._add_perm_impl('*', False)
        elif not perm.default_enable and '*' in users:
            # If a permission changed meanwhile (this should NOT be a usual
            # case!), do not let the '*' be locked into the database forever.
            handler._rem_perm_impl('*', False)


def require_permission(permission, extra_params, user):
    """
    Returns whether or not the given user has the given permission.

    :param extra_params: The scope-specific argument dict, which already
      contains a valid database session.
    """

    handler = handler_from_scope_params(permission,
                                        extra_params)
    if handler.has_permission(user):
        return True

    # If the user for some reason does not have the permission directly
    # (or by default), we need to walk the inheritance chain of the permission.
    ancestors = permission.inherited_from
    while ancestors:
        handler = handler_from_scope_params(ancestors[0], extra_params)

        if handler.has_permission(user):
            return True
        else:
            ancestors = ancestors[1:] + ancestors[0].inherited_from

    return False


def require_manager(permission, extra_params, user):
    """
    Returns whether or not the given user has rights to manage the given
    permission.

    :param extra_params: The scope-specific argument dict, which already
      contains a valid database session.
    """

    for manager in permission.managed_by:
        manager_handler = handler_from_scope_params(manager,
                                                    extra_params)
        if manager_handler.has_permission(user):
            return True

    return False

# ---------------------------------------------------------------------------
# Define the permissions available.
# Please refer to the user guide and the API documentation on which actions
# require which permission in particular.

# -- System-level permissions --


SUPERUSER = _create_permission(SystemPermission, 'SUPERUSER',
                               default_enable=False,
                               manages_self=True)

PERMISSION_VIEW = _create_permission(
    SystemPermission, 'PERMISSION_VIEW',
    default_enable=False,
    managed_by=[SUPERUSER])

# -- Product-level permissions --

PRODUCT_ADMIN = _create_permission(ProductPermission, 'PRODUCT_ADMIN',
                                   default_enable=False,
                                   managed_by=[SUPERUSER],
                                   manages_self=True)

PRODUCT_STORE = _create_permission(ProductPermission, 'PRODUCT_STORE',
                                   default_enable=False,
                                   inherited_from=[PRODUCT_ADMIN],
                                   managed_by=[PRODUCT_ADMIN])

PRODUCT_ACCESS = _create_permission(ProductPermission, 'PRODUCT_ACCESS',
                                    default_enable=False,
                                    inherited_from=[PRODUCT_ADMIN,
                                                    PRODUCT_STORE],
                                    managed_by=[PRODUCT_ADMIN])

PRODUCT_VIEW = _create_permission(ProductPermission, 'PRODUCT_VIEW',
                                  default_enable=False,
                                  inherited_from=[PRODUCT_ACCESS,
                                                  PRODUCT_ADMIN,
                                                  PRODUCT_STORE],
                                  managed_by=[PRODUCT_ADMIN])



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Defines the CodeChecker action for parsing a set of analysis results into a
human-readable format.
"""


import argparse
import os
import sys
from typing import Dict, Optional, Set
import fnmatch

from codechecker_report_converter.util import dump_json_output
from codechecker_report_converter.report import report_file, \
    reports as reports_helper
from codechecker_report_converter.report.output import baseline, codeclimate, \
    gerrit, json as report_to_json, plaintext
from codechecker_report_converter.report.output.html import \
    html as report_to_html
from codechecker_report_converter.report.statistics import Statistics


from codechecker_analyzer import analyzer_context, suppress_handler

from codechecker_common import arg, logger, cmd_config
from codechecker_common.review_status_handler import ReviewStatusHandler
from codechecker_common.skiplist_handler import SkipListHandler, \
    SkipListHandlers
from codechecker_common.source_code_comment_handler import \
    REVIEW_STATUS_VALUES
from codechecker_common.util import load_json


LOG = logger.get_logger('system')


def init_logger(level, stream=None, logger_name='system'):
    logger.setup_logger(level, stream)
    global LOG
    LOG = logger.get_logger(logger_name)


EXPORT_TYPES = ['html', 'json', 'codeclimate', 'gerrit', 'baseline']

epilog_env_var = f"""
  CC_CHANGED_FILES       Path of changed files json from Gerrit. Use it when
                         generating gerrit output.
  CC_REPO_DIR            Root directory of the sources, i.e. the directory
                         where the repository was cloned. Use it when
                         generating gerrit output.
  CC_REPORT_URL          URL where the report can be found. Use it when
                         generating gerrit output.
"""

epilog_exit_status = """
0 - No report
1 - CodeChecker error
2 - At least one report emitted by an analyzer
"""


def get_argparser_ctor_args():
    """
    This method returns a dict containing the kwargs for constructing an
    argparse.ArgumentParser (either directly or as a subparser).
    """
    return {
        'prog': 'CodeChecker parse',
        'formatter_class': arg.RawDescriptionDefaultHelpFormatter,

        # Description is shown when the command's help is queried directly
        'description': """
Parse and pretty-print the summary and results from one or more
'codechecker-analyze' result files. Bugs which are commented by using
"false_positive", "suppress" and "intentional" source code comments will not be
printed by the `parse` command.""",

        'epilog': f"""
Environment variables
------------------------------------------------
{epilog_env_var}

Exit status
------------------------------------------------
{epilog_exit_status}
""",

        # Help is shown when the "parent" CodeChecker command lists the
        # individual subcommands.
        'help': "Print analysis summary and results in a human-readable "
                "format."
    }


def add_arguments_to_parser(parser):
    """
    Add the subcommand's arguments to the given argparse.ArgumentParser.
    """

    parser.add_argument('input',
                        type=str,
                        nargs='+',
                        metavar='file/folder',
                        help="The analysis result files and/or folders "
                             "containing analysis results which should be "
                             "parsed and printed.")

    cmd_config.add_option(parser)

    parser.add_argument('-t', '--type', '--input-format',
                        dest="input_format",
                        required=False,
                        choices=['plist'],
                        default='plist',
                        help="Specify the format the analysis results were "
                             "created as.")

    output_opts = parser.add_argument_group("export arguments")
    output_opts.add_argument('-e', '--export',
                             dest="export",
                             required=False,
                             choices=EXPORT_TYPES,
                             help="R|Specify extra output format type.\n"
                                  "'codeclimate' format can be used for "
                                  "Code Climate and for GitLab integration. "
                                  "For more information see:\n"
                                  "https://github.com/codeclimate/platform/"
                                  "blob/master/spec/analyzers/SPEC.md"
                                  "#data-types\n"
                                  "'baseline' output can be used to integrate "
                                  "CodeChecker into your local workflow "
                                  "without using a CodeChecker server. For "
                                  "more information see our usage guide.")

    output_opts.add_argument('-o', '--output',
                             dest="output_path",
                             default=argparse.SUPPRESS,
                             help="Store the output in the given file/folder. "
                                  "Note: baseline files must have extension "
                                  "'.baseline'.")

    parser.add_argument('--suppress',
                        type=str,
                        dest="suppress",
                        default=argparse.SUPPRESS,
                        required=False,
                        help="Path of the suppress file to use. Records in "
                             "the suppress file are used to suppress the "
                             "display of certain results when parsing the "
                             "analyses' report. (Reports to an analysis "
                             "result can also be suppressed in the source "
                             "code -- please consult the manual on how to "
                             "do so.) NOTE: The suppress file relies on the "
                             "\"bug identifier\" generated by the analyzers "
                             "which is experimental, take care when relying "
                             "on it.")

    parser.add_argument('--export-source-suppress',
                        dest="create_suppress",
                        action="store_true",
                        required=False,
                        default=argparse.SUPPRESS,
                        help="Write suppress data from the suppression "
                             "annotations found in the source files that were "
                             "analyzed earlier that created the results. "
                             "The suppression information will be written "
                             "to the parameter of '--suppress'.")

    parser.add_argument('--print-steps',
                        dest="print_steps",
                        action="store_true",
                        required=False,
                        default=argparse.SUPPRESS,
                        help="Print the steps the analyzers took in finding "
                             "the reported defect.")

    parser.add_argument('--trim-path-prefix',
                        type=str,
                        nargs='*',
                        dest="trim_path_prefix",
                        required=False,
                        default=argparse.SUPPRESS,
                        help="Removes leading path from files which will be "
                             "printed. For instance if you analyze files "
                             "'/home/jsmith/my_proj/x.cpp' and "
                             "'/home/jsmith/my_proj/y.cpp', but would prefer "
                             "to have them displayed as 'my_proj/x.cpp' and "
                             "'my_proj/y.cpp' in the web/CLI interface, "
                             "invoke CodeChecker with '--trim-path-prefix "
                             "\"/home/jsmith/\"'."
                             "If multiple prefixes are given, the longest "
                             "match will be removed. You may also use Unix "
                             "shell-like wildcards (e.g. '/*/jsmith/').")

    parser.add_argument('--review-status',
                        nargs='*',
                        dest="review_status",
                        metavar='REVIEW_STATUS',
                        choices=REVIEW_STATUS_VALUES,
                        default=["confirmed", "unreviewed"],
                        help="Filter results by review statuses. Valid "
                             "values are: {0}".format(
                                 ', '.join(REVIEW_STATUS_VALUES)))

    group = parser.add_argument_group("file filter arguments")

    group.add_argument('-i', '--ignore', '--skip',
                       dest="skipfile",
                       required=False,
                       default=argparse.SUPPRESS,
                       help="Path to the Skipfile dictating which project "
                            "files should be omitted from analysis. Please "
                            "consult the User guide on how a Skipfile "
                            "should be laid out.")

    group.add_argument('--file',
                       nargs='+',
                       dest="files",
                       metavar='FILE',
                       required=False,
                       default=argparse.SUPPRESS,
                       help="Filter results by file path. "
                            "The file path can contain multiple * "
                            "quantifiers which matches any number of "
                            "characters (zero or more). So if you have "
                            "/a/x.cpp and /a/y.cpp then \"/a/*.cpp\" "
                            "selects both.")

    logger.add_verbose_arguments(parser)
    parser.set_defaults(
        func=main, func_process_config_file=cmd_config.process_config_file)


def ch_workdir(metadata: Optional[Dict]):
    """ Change working directory to the one noted in metadata.json if this
    file exists and contains "working_directory".
    """
    if not metadata or 'working_directory' not in metadata:
        return

    working_dir = metadata['working_directory']
    try:
        os.chdir(working_dir)
    except OSError as oerr:
        LOG.debug(oerr)
        LOG.error("Working directory %s is missing.\nCan not parse reports "
                  "safely.", working_dir)
        sys.exit(1)


def get_metadata(dir_path: str) -> Optional[Dict]:
    """ Get metadata from the given dir path or None if not exists. """
    metadata_file = os.path.join(dir_path, "metadata.json")
    if os.path.exists(metadata_file):
        return load_json(metadata_file)

    return None


def main(args):
    """
    Entry point for parsing some analysis results and printing them to the
    stdout in a human-readable format.
    """
    # If the given output format is not 'table', redirect logger's output to
    # the stderr.
    stream = None
    if 'export' in args and args.export not in [None, 'table', 'html']:
        stream = 'stderr'

    init_logger(args.verbose if 'verbose' in args else None, stream)

    try:
        cmd_config.check_config_file(args)
    except FileNotFoundError as fnerr:
        LOG.error(fnerr)
        sys.exit(1)

    incorrect_input_paths = [input_path for input_path in args.input
                             if not os.path.exists(input_path)]

    if incorrect_input_paths:
        for input_path in incorrect_input_paths:
            LOG.error(f"Input path {input_path} does not exist!")
        sys.exit(1)

    export = args.export if 'export' in args else None
    if export == 'html' and 'output_path' not in args:
        LOG.error("Argument --export not allowed without argument --output "
                  "when exporting to HTML.")
        sys.exit(1)

    if export == 'gerrit' and not gerrit.mandatory_env_var_is_set():
        sys.exit(1)

    if export and export not in EXPORT_TYPES:
        LOG.error("Unknown export format: %s", export)
        sys.exit(1)

    context = analyzer_context.get_context()

    # To ensure the help message prints the default folder properly,
    # the 'default' for 'args.input' is a string, not a list.
    # But we need lists for the foreach here to work.
    if isinstance(args.input, str):
        args.input = [args.input]

    src_comment_status_filter = args.review_status

    suppr_handler = None
    if 'suppress' in args:
        __make_handler = False
        if not os.path.isfile(args.suppress):
            if 'create_suppress' in args:
                with open(args.suppress, 'w',
                          encoding='utf-8', errors='ignore') as _:
                    # Just create the file.
                    __make_handler = True
                    LOG.info("Will write source-code suppressions to "
                             "suppress file: %s", args.suppress)
            else:
                LOG.warning("Suppress file '%s' given, but it does not exist"
                            " -- will not suppress anything.", args.suppress)
        else:
            __make_handler = True

        if __make_handler:
            suppr_handler = suppress_handler.\
                GenericSuppressHandler(args.suppress,
                                       'create_suppress' in args,
                                       src_comment_status_filter)
    elif 'create_suppress' in args:
        LOG.error("Can't use '--export-source-suppress' unless '--suppress "
                  "SUPPRESS_FILE' is also given.")
        sys.exit(1)

    output_dir_path = None
    output_file_path = None
    if 'output_path' in args:
        output_path = os.path.abspath(args.output_path)

        if export == 'html':
            output_dir_path = output_path
        else:
            if os.path.exists(output_path) and os.path.isdir(output_path):
                # For backward compatibility reason we handle the use case
                # when directory is provided to this command.
                LOG.error("Please provide a file path instead of a directory "
                          "for '%s' export type!", export)
                sys.exit(1)

            if export == 'baseline' and not baseline.check(output_path):
                LOG.error("Baseline files must have '.baseline' extensions.")
                sys.exit(1)

            output_file_path = output_path
            output_dir_path = os.path.dirname(output_file_path)

        if not os.path.exists(output_dir_path):
            os.makedirs(output_dir_path)

    def get_output_file_path(default_file_name: str) -> Optional[str]:
        """ Return an output file path. """
        if output_file_path:
            return output_file_path

        if output_dir_path:
            return os.path.join(output_dir_path, default_file_name)

    skip_handlers = SkipListHandlers()
    if 'files' in args:
        items = [f"+{file_path}" for file_path in args.files]
        items.append("-*")
        skip_handlers.append(SkipListHandler("\n".join(items)))
    if 'skipfile' in args:
        with open(args.skipfile, 'r', encoding='utf-8', errors='ignore') as f:
            skip_handlers.append(SkipListHandler(f.read()))

    trim_path_prefixes = args.trim_path_prefix if \
        'trim_path_prefix' in args else None

    all_reports = []
    statistics = Statistics()
    file_cache = {}  # For memory effiency.
    changed_files: Set[str] = set()
    processed_path_hashes = set()
    processed_file_paths = set()
    print_steps = 'print_steps' in args
    review_status_handler = ReviewStatusHandler()

    html_builder: Optional[report_to_html.HtmlBuilder] = None
    if export == 'html':
        html_builder = report_to_html.HtmlBuilder(
            context.path_plist_to_html_dist,
            context.checker_labels)

    for dir_path, file_paths in report_file.analyzer_result_files(args.input):
        review_status_cfg = os.path.join(dir_path, 'review_status.yaml')
        if os.path.lexists(review_status_cfg):
            try:
                review_status_handler.set_review_status_config(
                    review_status_cfg)
            except ValueError as err:
                LOG.error(err)
                sys.exit(1)

        metadata = get_metadata(dir_path)

        if metadata and 'files' in args:
            # Mapping plists when files are specified to speed up parsing
            # The specifed_file_paths variable would be an empty list
            # if metadata.json did not contain the specified file or
            # metadata did not contain mapping between source files and plists.
            specifed_file_paths = [
                key for key, val in
                metadata['tools'][0]['result_source_files'].items()
                if any(fnmatch.fnmatch(val, f) for f in args.files)
            ] if 'tools' in metadata \
                and len(metadata['tools']) > 0 \
                and 'result_source_files' in metadata['tools'][0] \
                else []
            file_paths = specifed_file_paths or file_paths

        for file_path in file_paths:
            reports = report_file.get_reports(
                file_path, context.checker_labels, file_cache)

            for report in reports:
                try:
                    # TODO: skip_handler is used later in reports_helper.skip()
                    # too. However, skipped reports shouldn't check source code
                    # comments because they potentially raise an exception.
                    # Skipped files shouldn't raise an exception, also, "skip"
                    # shouldn't be checked twice.
                    if not report.skip(skip_handlers):
                        report.review_status = \
                            review_status_handler.get_review_status(report)
                except ValueError as err:
                    LOG.error(err)
                    sys.exit(1)

            reports = reports_helper.skip(
                reports, processed_path_hashes, skip_handlers, suppr_handler,
                src_comment_status_filter)

            statistics.num_of_analyzer_result_files += 1
            for report in reports:
                if report.changed_files:
                    changed_files.update(report.changed_files)

                statistics.add_report(report)

                if trim_path_prefixes:
                    report.trim_path_prefixes(trim_path_prefixes)

            all_reports.extend(reports)

            # Print reports continously.
            if not export:
                file_report_map = plaintext.get_file_report_map(
                    reports, file_path, metadata)
                plaintext.convert(
                    review_status_handler,
                    file_report_map, processed_file_paths, print_steps)
            elif export == 'html':
                print(f"Parsing input file '{file_path}'.")
                report_to_html.convert(
                    file_path, reports, output_dir_path,
                    html_builder)

    for warning in review_status_handler.source_comment_warnings():
        LOG.warning(warning)

    if export is None:  # Plain text output
        statistics.write()
    elif export == 'html':
        html_builder.finish(output_dir_path, statistics)
    elif export == 'json':
        data = report_to_json.convert(all_reports)
        dump_json_output(data, get_output_file_path("reports.json"))
    elif export == 'codeclimate':
        data = codeclimate.convert(all_reports)
        dump_json_output(data, get_output_file_path("reports.json"))
    elif export == 'gerrit':
        data = gerrit.convert(all_reports)
        dump_json_output(data, get_output_file_path("reports.json"))
    elif export == 'baseline':
        data = baseline.convert(all_reports)
        output_path = get_output_file_path("reports.baseline")
        if output_path:
            baseline.write(output_path, data)

    reports_helper.dump_changed_files(changed_files)

    if statistics.num_of_reports:
        sys.exit(2)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

"""
Supported analyzer types.
"""


import os
import re
import subprocess
import sys

from codechecker_analyzer import analyzer_context
from codechecker_common.logger import get_logger

from .. import host_check

from .clangtidy.analyzer import ClangTidy
from .clangsa.analyzer import ClangSA
from .cppcheck.analyzer import Cppcheck
from .gcc.analyzer import Gcc

LOG = get_logger('analyzer')

supported_analyzers = {ClangSA.ANALYZER_NAME: ClangSA,
                       ClangTidy.ANALYZER_NAME: ClangTidy,
                       Cppcheck.ANALYZER_NAME: Cppcheck,
                       Gcc.ANALYZER_NAME: Gcc}


def is_ctu_capable():
    """ Detects if the current clang is CTU compatible. """
    enabled_analyzers, _ = check_supported_analyzers([ClangSA.ANALYZER_NAME])
    if not enabled_analyzers:
        return False

    return ClangSA.ctu_capability().is_ctu_capable


def is_ctu_on_demand_available():
    """ Detects if the current clang is capable of on-demand AST loading. """
    enabled_analyzers, _ = check_supported_analyzers([ClangSA.ANALYZER_NAME])
    if not enabled_analyzers:
        return False

    return ClangSA.ctu_capability().is_on_demand_ctu_available


def is_statistics_capable():
    """ Detects if the current clang is Statistics compatible. """
    # Resolve potentially missing binaries.
    enabled_analyzers, _ = check_supported_analyzers([ClangSA.ANALYZER_NAME])
    if not enabled_analyzers:
        return False

    checkers = ClangSA.get_analyzer_checkers(alpha=True, debug=True)

    stat_checkers_pattern = re.compile(r'.+statisticscollector.+')

    for checker_name, _ in checkers:
        if stat_checkers_pattern.match(checker_name):
            return True

    return False


def is_z3_capable():
    """ Detects if the current clang is Z3 compatible. """
    enabled_analyzers, _ = check_supported_analyzers([ClangSA.ANALYZER_NAME])
    if not enabled_analyzers:
        return False

    return host_check.has_analyzer_option(ClangSA.analyzer_binary(),
                                          ['-Xclang',
                                           '-analyzer-constraints=z3'])


def is_z3_refutation_capable():
    """ Detects if the current clang is Z3 refutation compatible. """

    # This function basically checks whether the corresponding analyzer config
    # option exists i.e. it is visible on analyzer config option help page.
    # However, it doesn't mean that Clang itself is compiled with Z3.
    if not is_z3_capable():
        return False

    check_supported_analyzers([ClangSA.ANALYZER_NAME])

    return host_check.has_analyzer_config_option(ClangSA.analyzer_binary(),
                                                 'crosscheck-with-z3')


def is_ignore_conflict_supported():
    """
    Detects if clang-apply-replacements supports --ignore-insert-conflict flag.
    """
    context = analyzer_context.get_context()
    proc = subprocess.Popen([context.replacer_binary, '--help'],
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            env=context.analyzer_env,
                            encoding="utf-8", errors="ignore")
    out, _ = proc.communicate()
    return '--ignore-insert-conflict' in out


def print_unsupported_analyzers(errored):
    """ Print error messages which occured during analyzer detection. """
    for analyzer_binary, reason in errored:
        LOG.warning("Analyzer '%s' is enabled but CodeChecker is failed to "
                    "execute analysis with it: '%s'. Please check your "
                    "'PATH' environment variable and the "
                    "'config/package_layout.json' file!",
                    analyzer_binary, reason)


def check_available_analyzers(analyzers, errored):
    """ Handle use case when no analyzer can be found on the user machine. """
    if analyzers:
        return

    print_unsupported_analyzers(errored)
    LOG.error("Failed to run command because no analyzers can be found on "
              "your machine!")
    sys.exit(1)


def check_supported_analyzers(analyzers):
    """
    Checks the given analyzers in the current context for their executability
    and support in CodeChecker.

    This method also updates the given context.analyzer_binaries if the
    context's configuration is bogus but had been resolved.

    :return: (enabled, failed) where enabled is a list of analyzer names
     and failed is a list of (analyzer, reason) tuple.
    """

    context = analyzer_context.get_context()
    check_env = context.analyzer_env

    analyzer_binaries = context.analyzer_binaries

    enabled_analyzers = set()
    failed_analyzers = set()

    for analyzer_name in analyzers:
        if analyzer_name not in supported_analyzers:
            failed_analyzers.add((analyzer_name,
                                  "Analyzer unsupported by CodeChecker!"))
            continue

        # Get the compiler binary to check if it can run.
        available_analyzer = True
        analyzer_bin = analyzer_binaries.get(analyzer_name)
        if not analyzer_bin:
            failed_analyzers.add((analyzer_name,
                                  "Failed to detect analyzer binary!"))
            continue
        elif not os.path.isabs(analyzer_bin):
            # If the analyzer is not in an absolute path, try to find it...
            found_bin = supported_analyzers[analyzer_name].\
                resolve_missing_binary(analyzer_bin, check_env)

            # found_bin is an absolute path, an executable in one of the
            # PATH folders.
            # If found_bin is the same as the original binary, ie., normally
            # calling the binary without any search would have resulted in
            # the same binary being called, it's NOT a "not found".
            if found_bin and os.path.basename(found_bin) != analyzer_bin:
                LOG.debug("Configured binary '%s' for analyzer '%s' was "
                          "not found, but environment PATH contains '%s'.",
                          analyzer_bin, analyzer_name, found_bin)
                context.analyzer_binaries[analyzer_name] = \
                    os.path.realpath(found_bin)

            analyzer_bin = found_bin

        # Check version compatibility of the analyzer binary.
        if analyzer_bin:
            analyzer = supported_analyzers[analyzer_name]
            error = analyzer.is_binary_version_incompatible(check_env)
            if error:
                failed_analyzers.add((analyzer_name,
                                      f"Incompatible version: {error} "
                                      "Maybe try setting an absolute path to "
                                      "a different analyzer binary via the "
                                      "env variable CC_ANALYZER_BIN?"))
                available_analyzer = False

        if not analyzer_bin or \
           not host_check.check_analyzer(analyzer_bin, check_env):
            # Analyzers unavailable under absolute paths are deliberately a
            # configuration problem.
            failed_analyzers.add((analyzer_name,
                                  "Cannot execute analyzer binary!"))
            available_analyzer = False

        if available_analyzer:
            enabled_analyzers.add(analyzer_name)

    return enabled_analyzers, failed_analyzers


def construct_analyzer(buildaction,
                       analyzer_config):
    try:
        analyzer_type = buildaction.analyzer_type

        LOG.debug_analyzer('Constructing %s analyzer.', analyzer_type)
        if analyzer_type in supported_analyzers:
            analyzer = supported_analyzers[analyzer_type](analyzer_config,
                                                          buildaction)
        else:
            analyzer = None
            LOG.error('Unsupported analyzer type: %s', analyzer_type)
        return analyzer

    except Exception:
        # We should've detected well before this point that something is off
        # with the analyzer. We can't recover here.
        raise


def build_config_handlers(args, enabled_analyzers):
    """
    Handle config from command line or from config file if no command line
    config is given.

    Supported command line config format is in JSON tidy supports YAML also but
    no standard lib for yaml parsing is available in python.
    """

    analyzer_config_map = {}

    for ea in enabled_analyzers:
        assert supported_analyzers[ea].analyzer_binary(), \
            "At this point, we should've checked if this analyzer has a " \
            "binary!"
        config_handler = supported_analyzers[ea].construct_config_handler(args)
        analyzer_config_map[ea] = config_handler

    return analyzer_config_map
